from optparse import make_option

from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS
import django

from denorm import denorms


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument(
            '--database', action='store', dest='database',
            default=DEFAULT_DB_ALIAS, help='Nominates a database to execute '
                'SQL into. Defaults to the "default" database.'
        ),

    help = "Removes all triggers created by django-denorm."

    def handle(self, **options):
        using = options['database']
        denorms.drop_triggers(using=using)



#!/usr/bin/python
import os
import sys

dbtypes = ["postgres"]

os.environ["PYTHONPATH"] = ".:..:test_denorm_project:../denorm:test_app"

for dbtype in dbtypes:
    print("running tests on %s" % dbtype)
    os.environ["DJANGO_SETTINGS_MODULE"] = "test_denorm_project.settings_%s" % dbtype

    test_label = sys.argv[2] if len(sys.argv) > 2 else "test_app"
    if os.system(
        "cd test_denorm_project; coverage run manage.py test --noinput --keepdb %s"
        % test_label
    ):
        exit(1)



# Generated by Django 3.2.7 on 2021-10-03 03:37

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("denorm", "0004_alter_dirtyinstance_success"),
    ]

    operations = [
        migrations.AddField(
            model_name="dirtyinstance",
            name="created_on",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 3.2.7 on 2021-10-02 20:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("denorm", "0003_auto_20211002_1955"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dirtyinstance",
            name="success",
            field=models.NullBooleanField(),
        ),
    ]



import contextlib


@contextlib.contextmanager
def suppress_autotime(model, fields):
    _original_values = {}
    for field in model._meta.local_fields:
        if field.name in fields:
            _original_values[field.name] = {
                "auto_now": field.auto_now,
                "auto_now_add": field.auto_now_add,
            }
            field.auto_now = False
            field.auto_now_add = False
    try:
        yield
    finally:
        for field in model._meta.local_fields:
            if field.name in fields:
                field.auto_now = _original_values[field.name]["auto_now"]
                field.auto_now_add = _original_values[field.name]["auto_now_add"]



# Generated by Django 4.2.3 on 2023-07-09 20:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("denorm", "0001_squashed_0012_alter_dirtyinstance_object_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dirtyinstance",
            name="success",
            field=models.BooleanField(db_index=True, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models

from denorm.db import const


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="DirtyInstance",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("object_id", models.TextField(null=True, blank=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        to="contenttypes.ContentType", on_delete=models.CASCADE
                    ),
                ),
            ],
        ),
        migrations.RunSQL(
            f"""
        -- Notify when record get inserted into 'django_denorm' table
        CREATE OR REPLACE FUNCTION notify_django_denorm_queue()
          RETURNS trigger AS $$
        DECLARE
        BEGIN
          PERFORM pg_notify('{const.DENORM_QUEUE_NAME}', '');
          RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;

        CREATE TRIGGER notify_django_denorm_queue
          AFTER INSERT ON denorm_dirtyinstance
          FOR EACH STATEMENT
          EXECUTE PROCEDURE notify_django_denorm_queue();
        """,
            """
            DROP TRIGGER notify_django_denorm_queue ON denorm_dirtyinstance;
            DROP FUNCTION notify_django_denorm_queue;
            """,
        ),
    ]



import os

from test_denorm_project.settings import *  # noqa

TOX_ENVIRONMENT = os.getenv("TOX_PARALLEL_ENV")
DB_SUFFIX = ""
if TOX_ENVIRONMENT:
    DB_SUFFIX = "_" + TOX_ENVIRONMENT

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": f"denorm_test{DB_SUFFIX}",
        "HOST": "localhost",
        "USER": "postgres",
        "PASSWORD": "",
    }
}



# Generated by Django 3.2.7 on 2021-10-03 03:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("denorm", "0006_auto_20211003_0346"),
    ]

    operations = [
        migrations.RunSQL(
            "ALTER TABLE denorm_dirtyinstance ALTER COLUMN created_on SET DEFAULT CURRENT_TIMESTAMP",
            migrations.RunSQL.noop,
        )
    ]



# Generated by Django 5.2.5 on 2025-08-18 11:45

from django.db import migrations

from denorm.db import const


class Migration(migrations.Migration):
    dependencies = [
        ("denorm", "0013_alter_dirtyinstance_success"),
    ]

    operations = [
        migrations.RunSQL(
            f"""
    -- Notify when record get inserted into 'django_denorm' table
    CREATE OR REPLACE FUNCTION notify_django_denorm_queue()
      RETURNS trigger AS $$
    DECLARE
    BEGIN
      PERFORM pg_notify('{const.DENORM_QUEUE_NAME}', NEW.id::text);
      RETURN NEW;
    END;
    $$ LANGUAGE plpgsql;


    CREATE OR REPLACE TRIGGER notify_django_denorm_queue
      AFTER INSERT ON denorm_dirtyinstance
      FOR EACH ROW
      EXECUTE PROCEDURE notify_django_denorm_queue();
      """,
            migrations.RunSQL.noop,
        )
    ]



from io import StringIO
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.management import call_command
from django.db import connection
from django.test import TestCase, TransactionTestCase
from test_app import models

import denorm
from denorm import denorms

User = get_user_model()


# Use all but denorms in FailingTriggers models by default
failingdenorms = denorms.get_alldenorms()
denorms.alldenorms = [
    d
    for d in failingdenorms
    if d.model not in (models.FailingTriggersModelA, models.FailingTriggersModelB)
]


class TestTriggers(TestCase):
    def setUp(self):
        denorms.drop_triggers()

    def test_triggers(self):
        """Test potentially failing denorms."""
        # save and restore alldenorms
        # test will fail if it's raising an exception
        alldenorms = denorms.get_alldenorms()
        denorms.alldenorms = failingdenorms
        try:
            denorms.install_triggers()
        finally:
            denorms.alldenorms = alldenorms


class TestCached(TestCase):
    def setUp(self):
        denorms.drop_triggers()
        denorms.install_triggers()

    def tearDown(self):
        models.CachedModelA.objects.all().delete()
        models.CachedModelB.objects.all().delete()

    def test_depends_related(self):
        models.CachedModelB.objects.create(data="Hello")
        b = models.CachedModelB.objects.all()[0]
        self.assertEqual("Hello", b.data)

        models.CachedModelA.objects.create(b=b)
        a = models.CachedModelA.objects.all()[0]

        self.assertEqual("HELLO", a.cached_data["upper"])
        self.assertEqual("hello", a.cached_data["lower"])

        b.data = "World"
        self.assertEqual("HELLO", a.cached_data["upper"])
        self.assertEqual("hello", a.cached_data["lower"])

        b.save()
        a = models.CachedModelA.objects.all()[0]
        self.assertEqual("WORLD", a.cached_data["upper"])
        self.assertEqual("world", a.cached_data["lower"])


class TestAbstract(TestCase):
    def setUp(self):
        denorms.drop_triggers()
        denorms.install_triggers()

    def test_abstract(self):
        d1 = models.RealDenormModel.objects.create(text="onion")
        self.assertEqual("Ham and onion", d1.ham)
        self.assertEqual("Eggs and onion", d1.eggs)


class TestSkip(TransactionTestCase):
    """
    Tests for the skip feature.
    """

    def setUp(self):
        denorms.drop_triggers()
        denorm.models.DirtyInstance.objects.all().delete()
        denorms.install_triggers()

        post = models.SkipPost(text="Here be ponies.")
        post.save()

        self.post = post

    # TODO: Enable and check!
    # Unsure on how to test this behaviour. It results in an endless loop:
    # update -> trigger -> update -> trigger -> ...
    #
    # def test_without_skip(self):
    #    # This results in an infinate loop on SQLite.
    #    comment = SkipCommentWithoutSkip(post=self.post, text='Oh really?')
    #    comment.save()
    #
    #    denorm.flush()

    # TODO: Check if an infinate loop happens and stop it.
    def test_with_skip(self):
        # This should not result in an endless loop.
        comment = models.SkipCommentWithSkip(post=self.post, text="Oh really?")
        comment.save()

        denorm.flush()

    def test_with_only(self):
        # This should not result in an endless loop.
        comment = models.SkipCommentWithOnly(post=self.post, text="Oh really?")
        comment.save()

        denorm.flush()

    def test_meta_skip(self):
        """Test a model with the attribute listed under denorm_always_skip."""
        comment = models.SkipCommentWithAttributeSkip(
            post=self.post, text="Yup, and they have wings!"
        )
        comment.save()

        denorm.flush()


class TestDirtyInstance(TestCase):
    def test___str__(self):
        from test_app.models import CachedModelB

        from denorm.models import DirtyInstance

        x = CachedModelB.objects.create(data="Hello")
        d = DirtyInstance(content_object=x, func_name=None)
        assert str(d).find("DirtyInstance") >= 0


class TestDenormalisation(TransactionTestCase):
    """
    Tests for the denormalisation fields.
    """

    def setUp(self):
        denorms.drop_triggers()
        denorm.models.DirtyInstance.objects.all().delete()
        denorms.install_triggers()

        self.testuser = User.objects.create_user("testuser", "testuser", "testuser")
        self.testuser.is_staff = True
        ContentType.objects.get_for_model(models.Member)
        self.testuser.save()

    def tearDown(self):
        # delete all model instances
        self.testuser.delete()
        models.Attachment.objects.all().delete()
        models.Post.objects.all().delete()
        models.Forum.objects.all().delete()

    def test_depends_related(self):
        """
        test_depends_related -- Test the DependsOnRelated stuff.
        """
        # Make a forum, check it's got no posts
        f1 = models.Forum.objects.create(title="forumone")
        self.assertEqual(f1.post_count, 0)
        # Check its database copy too
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 0)

        # Add a post
        p1 = models.Post.objects.create(forum=f1)
        # Has the post count updated?
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 1)
        denorm.flush()

        # Check its title, in p1 and the DB
        self.assertEqual(p1.forum_title, "forumone")
        self.assertEqual(models.Post.objects.get(id=p1.id).forum_title, "forumone")

        # Update the forum title
        f1.title = "forumtwo"
        f1.save()

        denorm.flush()

        # Has the post's title changed?
        self.assertEqual(models.Post.objects.get(id=p1.id).forum_title, "forumtwo")

        # Add and remove some posts and check the post count
        models.Post.objects.create(forum=f1)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 2)
        models.Post.objects.create(forum=f1)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 3)
        p1.delete()
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 2)

        # Delete everything, check once more.
        models.Post.objects.all().delete()
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 0)

        # Make an orphaned post, see what its title is.
        # Doesn't work yet - no support for null FKs
        # p4 = Post.objects.create(forum=None)
        # self.assertEqual(p4.forum_title, None)

    def test_dependency_chains(self):
        # create a forum, a member and a post
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(name="memberone")
        models.Post.objects.create(forum=f1, author=m1)
        denorm.flush()

        # check the forums author list contains the member
        self.assertEqual(models.Forum.objects.get(id=f1.id).author_names, "memberone")

        # change the member's name
        m1.name = "membertwo"
        m1.save()
        denorm.flush()

        # check again
        self.assertEqual(models.Forum.objects.get(id=f1.id).author_names, "membertwo")

    def test_trees(self):
        f1 = models.Forum.objects.create(title="forumone")
        f2 = models.Forum.objects.create(title="forumtwo", parent_forum=f1)
        f3 = models.Forum.objects.create(title="forumthree", parent_forum=f2)
        denorm.flush()

        self.assertEqual(f1.path, "/forumone/")
        self.assertEqual(f2.path, "/forumone/forumtwo/")
        self.assertEqual(f3.path, "/forumone/forumtwo/forumthree/")

        f1.title = "someothertitle"
        f1.save()
        denorm.flush()

        f1 = models.Forum.objects.get(id=f1.id)
        f2 = models.Forum.objects.get(id=f2.id)
        f3 = models.Forum.objects.get(id=f3.id)

        self.assertEqual(f1.path, "/someothertitle/")
        self.assertEqual(f2.path, "/someothertitle/forumtwo/")
        self.assertEqual(f3.path, "/someothertitle/forumtwo/forumthree/")

    def test_reverse_fk_null(self):
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(name="memberone")
        models.Post.objects.create(forum=f1, author=m1)
        models.Attachment.objects.create()
        denorm.flush()

    def test_bulk_update(self):
        """
        test_bulk_update -- Test the DependsOnRelated stuff.
        """
        f1 = models.Forum.objects.create(title="forumone")
        f2 = models.Forum.objects.create(title="forumtwo")
        p1 = models.Post.objects.create(forum=f1)
        p2 = models.Post.objects.create(forum=f2)
        # denorms.INTERACTIVE = True
        denorm.flush()

        self.assertEqual(models.Post.objects.get(id=p1.id).forum_title, "forumone")
        self.assertEqual(models.Post.objects.get(id=p2.id).forum_title, "forumtwo")
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 1)
        self.assertEqual(models.Forum.objects.get(id=f2.id).post_count, 1)

        models.Post.objects.update(forum=f1)
        denorm.flush()
        self.assertEqual(models.Post.objects.get(id=p1.id).forum_title, "forumone")
        self.assertEqual(models.Post.objects.get(id=p2.id).forum_title, "forumone")
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 2)
        self.assertEqual(models.Forum.objects.get(id=f2.id).post_count, 0)

        models.Forum.objects.update(title="oneforall")
        denorm.flush()
        self.assertEqual(models.Post.objects.get(id=p1.id).forum_title, "oneforall")
        self.assertEqual(models.Post.objects.get(id=p2.id).forum_title, "oneforall")

    def test_no_dependency(self):
        m1 = models.Member.objects.create(first_name="first", name="last")
        denorm.flush()

        self.assertEqual(models.Member.objects.get(id=m1.id).full_name, "first last")

        models.Member.objects.filter(id=m1.id).update(first_name="second")
        denorm.flush()
        self.assertEqual(models.Member.objects.get(id=m1.id).full_name, "second last")

    def test_self_backward_relation(self):

        f1 = models.Forum.objects.create(title="forumone")
        p1 = models.Post.objects.create(
            forum=f1,
        )
        p2 = models.Post.objects.create(forum=f1, response_to=p1)
        p3 = models.Post.objects.create(forum=f1, response_to=p1)
        p4 = models.Post.objects.create(forum=f1, response_to=p2)
        denorm.flush()

        self.assertEqual(models.Post.objects.get(id=p1.id).response_count, 3)
        self.assertEqual(models.Post.objects.get(id=p2.id).response_count, 1)
        self.assertEqual(models.Post.objects.get(id=p3.id).response_count, 0)
        self.assertEqual(models.Post.objects.get(id=p4.id).response_count, 0)

    def test_m2m_relation(self):
        f1 = models.Forum.objects.create(title="forumone")
        p1 = models.Post.objects.create(forum=f1, title="post1")
        m1 = models.Member.objects.create(first_name="first1", name="last1")

        denorm.flush()
        m1.bookmarks.add(p1)
        denorm.flush()

        self.assertTrue("post1" in models.Member.objects.get(id=m1.id).bookmark_titles)
        p1.title = "othertitle"
        p1.save()
        denorm.flush()
        self.assertTrue(
            "post1" not in models.Member.objects.get(id=m1.id).bookmark_titles
        )
        self.assertTrue(
            "othertitle" in models.Member.objects.get(id=m1.id).bookmark_titles
        )

        p2 = models.Post.objects.create(forum=f1, title="thirdtitle")
        m1.bookmarks.add(p2)
        denorm.flush()
        self.assertTrue(
            "post1" not in models.Member.objects.get(id=m1.id).bookmark_titles
        )
        self.assertTrue(
            "othertitle" in models.Member.objects.get(id=m1.id).bookmark_titles
        )
        self.assertTrue(
            "thirdtitle" in models.Member.objects.get(id=m1.id).bookmark_titles
        )

        m1.bookmarks.remove(p1)
        denorm.flush()
        self.assertTrue(
            "othertitle" not in models.Member.objects.get(id=m1.id).bookmark_titles
        )
        self.assertTrue(
            "thirdtitle" in models.Member.objects.get(id=m1.id).bookmark_titles
        )

    def test_middleware(self):
        # FIXME, this test currently does not work with a transactional
        # database, so it's skipped for now.
        return
        # FIXME, set and de-set middleware values
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(first_name="first1", name="last1")
        p1 = models.Post.objects.create(forum=f1, author=m1)

        self.assertEqual(models.Post.objects.get(id=p1.id).author_name, "last1")

        self.client.login(username="testuser", password="testuser")
        self.client.post(
            "/admin/denorm_testapp/member/%s/" % (m1.pk),
            {
                "name": "last2",
                "first_name": "first2",
            },
        )

        self.assertEqual(models.Post.objects.get(id=p1.id).author_name, "last2")

    def test_countfield(self):
        f1 = models.Forum.objects.create(title="forumone")
        f2 = models.Forum.objects.create(title="forumone")
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 0)
        self.assertEqual(models.Forum.objects.get(id=f2.id).post_count, 0)

        models.Post.objects.create(forum=f1)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 1)
        self.assertEqual(models.Forum.objects.get(id=f2.id).post_count, 0)

        p2 = models.Post.objects.create(forum=f2)
        p3 = models.Post.objects.create(forum=f2)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 1)
        self.assertEqual(models.Forum.objects.get(id=f2.id).post_count, 2)

        p2.forum = f1
        p2.save()
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 2)
        self.assertEqual(models.Forum.objects.get(id=f2.id).post_count, 1)

        models.Post.objects.filter(pk=p3.pk).update(forum=f1)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 3)
        self.assertEqual(models.Forum.objects.get(id=f2.id).post_count, 0)

    def test_countfield_does_not_write_stale_value(self):
        f1 = models.Forum.objects.create(title="forumone")
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 0)
        models.Post.objects.create(forum=f1)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 1)

        f1 = models.Forum.objects.get(title="forumone")
        models.Post.objects.create(forum_id=f1.id)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 2)
        f1.title = "new"
        self.assertEqual(f1.post_count, 1)
        f1.save()
        self.assertEqual(f1.post_count, 2)
        self.assertEqual(models.Forum.objects.get(id=f1.id).post_count, 2)
        self.assertEqual(models.Forum.objects.get(id=f1.id).title, "new")

    def test_foreignkey(self):
        f1 = models.Forum.objects.create(title="forumone")
        f2 = models.Forum.objects.create(title="forumtwo")
        m1 = models.Member.objects.create(first_name="first1", name="last1")
        p1 = models.Post.objects.create(forum=f1, author=m1)

        a1 = models.Attachment.objects.create(post=p1)
        self.assertEqual(models.Attachment.objects.get(id=a1.id).forum, f1)

        a2 = models.Attachment.objects.create()
        self.assertEqual(models.Attachment.objects.get(id=a2.id).forum, None)

        # Change forum
        p1.forum = f2
        p1.save()
        denorm.flush()
        self.assertEqual(models.Attachment.objects.get(id=a1.id).forum, f2)

        # test denorm function returning object, not PK
        models.Attachment.forum_as_object = True
        models.Attachment.objects.create(post=p1)
        models.Attachment.forum_as_object = False

    def test_m2m(self):
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(name="memberone")
        models.Post.objects.create(forum=f1, author=m1)
        denorm.flush()

        # check the forums author list contains the member
        self.assertTrue(m1 in models.Forum.objects.get(id=f1.id).authors.all())

        m2 = models.Member.objects.create(name="membertwo")
        p2 = models.Post.objects.create(forum=f1, author=m2)
        denorm.flush()

        self.assertTrue(m1 in models.Forum.objects.get(id=f1.id).authors.all())
        self.assertTrue(m2 in models.Forum.objects.get(id=f1.id).authors.all())

        p2.delete()
        denorm.flush()

        self.assertTrue(m2 not in models.Forum.objects.get(id=f1.id).authors.all())

    def test_denorm_rebuild(self):
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(name="memberone")
        p1 = models.Post.objects.create(forum=f1, author=m1)

        denorm.denorms.rebuildall()

        f1 = models.Forum.objects.get(id=f1.id)
        m1 = models.Member.objects.get(id=m1.id)
        p1 = models.Post.objects.get(id=p1.id)

        self.assertEqual(f1.post_count, 1)
        self.assertEqual(f1.authors.all()[0], m1)

    def test_denorm_rebuild_called_once(self):
        """
        Test whether the denorm function is not called only once during rebuild.
        The proxy model CallCounterProxy should not add extra callings to denorm function.
        """
        models.CallCounter.objects.create()
        denorm.denorms.flush()
        c = models.CallCounter.objects.get()
        # TODO: we should be able to create object with hitting the denorm function just once
        self.assertEqual(c.called_count, 2)
        denorm.denorms.rebuildall(verbose=True)
        c = models.CallCounter.objects.get()
        self.assertEqual(c.called_count, 3)
        denorm.denorms.rebuildall(verbose=True)
        c = models.CallCounter.objects.get()
        self.assertEqual(c.called_count, 4)

    def test_denorm_update(self):
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(name="memberone")
        p1 = models.Post.objects.create(forum=f1, author=m1)
        models.Attachment.objects.create(post=p1)

        denorm.denorms.rebuildall()

        f2 = models.Forum.objects.create(title="forumtwo")
        p1.forum = f2
        p1.save()

        # BUG https://github.com/initcrash/django-denorm/issues/24
        # We have to update the Attachment.forum field first to trigger this bug. Simply doing rebuildall() will
        # trigger an a1.save() at an some earlier point during the update. By the time we get to updating the value of
        # forum field the value is already correct and no update is done bypassing the broken code.
        denorm.denorms.rebuildall(model_name="Attachment", field_name="forum")

    def test_denorm_subclass(self):
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(name="memberone")
        p1 = models.Post.objects.create(forum=f1, author=m1)

        self.assertEqual(f1.tags_string, "")
        self.assertEqual(p1.tags_string, "")

        models.Tag.objects.create(name="tagone", content_object=f1)
        models.Tag.objects.create(name="tagtwo", content_object=f1)

        denorm.denorms.flush()
        f1 = models.Forum.objects.get(id=f1.id)
        m1 = models.Member.objects.get(id=m1.id)
        p1 = models.Post.objects.get(id=p1.id)

        self.assertEqual(f1.tags_string, "tagone, tagtwo")
        self.assertEqual(p1.tags_string, "")

        models.Tag.objects.create(name="tagthree", content_object=p1)
        t4 = models.Tag.objects.create(name="tagfour", content_object=p1)

        denorm.denorms.flush()
        f1 = models.Forum.objects.get(id=f1.id)
        m1 = models.Member.objects.get(id=m1.id)
        p1 = models.Post.objects.get(id=p1.id)

        self.assertEqual(f1.tags_string, "tagone, tagtwo")
        self.assertEqual(p1.tags_string, "tagfour, tagthree")

        t4.content_object = f1
        t4.save()

        denorm.denorms.flush()
        f1 = models.Forum.objects.get(id=f1.id)
        m1 = models.Member.objects.get(id=m1.id)
        p1 = models.Post.objects.get(id=p1.id)

        self.assertEqual(f1.tags_string, "tagfour, tagone, tagtwo")
        self.assertEqual(p1.tags_string, "tagthree")

    def test_denorm_delete(self):
        """This tests bug #69"""
        team = models.Team.objects.create()

        self.assertEqual(team.user_string, "")

        models.Competitor.objects.create(name="tagone", team=team)
        models.Competitor.objects.create(name="tagtwo", team=team)

        denorm.denorms.flush()
        team = models.Team.objects.get(id=team.id)
        self.assertEqual(team.user_string, "tagone, tagtwo")

        models.Competitor.objects.get(name="tagtwo").delete()

        denorm.denorms.flush()
        team = models.Team.objects.get(id=team.id)
        self.assertEqual(team.user_string, "tagone")

    def test_cache_key_field_backward(self):
        f1 = models.Forum.objects.create(title="forumone")
        f2 = models.Forum.objects.create(title="forumtwo")
        ck1 = f1.cachekey
        ck2 = f2.cachekey

        p1 = models.Post.objects.create(forum=f1)
        f1 = models.Forum.objects.get(id=f1.id)
        f2 = models.Forum.objects.get(id=f2.id)
        self.assertNotEqual(ck1, f1.cachekey)
        self.assertEqual(ck2, f2.cachekey)

        ck1 = f1.cachekey
        ck2 = f2.cachekey

        p1 = models.Post.objects.get(id=p1.id)
        p1.forum = f2
        p1.save()

        f1 = models.Forum.objects.get(id=f1.id)
        f2 = models.Forum.objects.get(id=f2.id)

        self.assertNotEqual(ck1, f1.cachekey)
        self.assertNotEqual(ck2, f2.cachekey)

    def test_cache_key_field_forward(self):
        f1 = models.Forum.objects.create(title="forumone")
        p1 = models.Post.objects.create(title="initial_title", forum=f1)
        a1 = models.Attachment.objects.create(post=p1)
        a2 = models.Attachment.objects.create(post=p1)

        a1 = models.Attachment.objects.get(id=a1.id)
        a2 = models.Attachment.objects.get(id=a2.id)
        self.assertNotEqual(a1.cachekey, a2.cachekey)

        ck1 = a1.cachekey
        ck2 = a2.cachekey
        p1.title = "new_title"
        p1.save()

        a1 = models.Attachment.objects.get(id=a1.id)
        a2 = models.Attachment.objects.get(id=a2.id)
        self.assertNotEqual(ck1, a1.cachekey)
        self.assertNotEqual(ck2, a2.cachekey)

        a1 = models.Attachment.objects.get(id=a1.id)
        a2 = models.Attachment.objects.get(id=a2.id)
        self.assertNotEqual(a1.cachekey, a2.cachekey)

    def test_cache_key_field_m2m(self):
        f1 = models.Forum.objects.create(title="forumone")
        m1 = models.Member.objects.create(name="memberone")
        p1 = models.Post.objects.create(title="initial_title", forum=f1)

        m1 = models.Member.objects.get(id=m1.id)
        ck1 = m1.cachekey

        m1.bookmarks.add(p1)

        m1 = models.Member.objects.get(id=m1.id)
        self.assertNotEqual(ck1, m1.cachekey)

        ck1 = m1.cachekey

        p1 = models.Post.objects.get(id=p1.id)
        p1.title = "new_title"
        p1.save()

        m1 = models.Member.objects.get(id=m1.id)
        self.assertNotEqual(ck1, m1.cachekey)


if connection.vendor != "sqlite":

    class TestFilterCount(TestCase):
        """
        Tests for the filtered count feature.
        """

        def setUp(self):
            denorms.drop_triggers()
            denorms.install_triggers()

        def test_filter_count(self):
            master = models.FilterCountModel.objects.create()
            self.assertEqual(master.active_item_count, 0)
            master.items.create(active=True, text="text")
            master.items.create(active=True, text="")
            master = models.FilterCountModel.objects.get(id=master.id)
            self.assertEqual(master.active_item_count, 1, "created active item")
            master.items.create(active=False)
            master = models.FilterCountModel.objects.get(id=master.id)
            self.assertEqual(master.active_item_count, 1, "created inactive item")
            master.items.create(active=True, text="true")
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 2)
            master.items.filter(active=False).delete()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 2)
            master.items.filter(active=True, text="true")[0].delete()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 1)
            item = master.items.filter(active=True, text="text")[0]
            item.active = False
            item.save()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 0)
            item = master.items.filter(active=False, text="text")[0]
            item.active = True
            item.text = ""
            item.save()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 0)
            item = master.items.filter(active=True, text="")[0]
            item.text = "123"
            item.save()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 1)

    class TestFilterCountM2M(TestCase):
        """
        Tests for the filtered count feature.
        """

        def setUp(self):
            denorms.drop_triggers()
            denorms.install_triggers()

        def test_filter_count(self):
            master = models.FilterCountModel.objects.create()
            self.assertEqual(master.active_item_count, 0)
            master.items.create(active=True, text="true")
            master = models.FilterCountModel.objects.get(id=master.id)
            self.assertEqual(master.active_item_count, 1, "created active item")
            master.items.create(active=False, text="true")
            master = models.FilterCountModel.objects.get(id=master.id)
            self.assertEqual(master.active_item_count, 1, "created inactive item")
            master.items.create(active=True, text="true")
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 2)
            master.items.filter(active=False).delete()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 2)
            master.items.filter(active=True)[0].delete()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 1)
            item = master.items.filter(active=True)[0]
            item.active = False
            item.save()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 0)
            item = master.items.filter(active=False)[0]
            item.active = True
            item.save()
            master = models.FilterCountModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_count, 1)

    class TestFilterSum(TestCase):
        """
        Tests for the filtered count feature.
        """

        def setUp(self):
            denorms.drop_triggers()
            denorms.install_triggers()

        def test_filter_count(self):
            master = models.FilterSumModel.objects.create()
            self.assertEqual(master.active_item_sum, 0)
            master.counts.create(age=18, active_item_count=8)
            master = models.FilterSumModel.objects.get(id=master.id)
            self.assertEqual(master.active_item_sum, 8)
            master.counts.create(age=16, active_item_count=10)
            master = models.FilterSumModel.objects.get(id=master.id)
            self.assertEqual(master.active_item_sum, 8, "created inactive item")
            master.counts.create(age=19, active_item_count=9)
            master = models.FilterSumModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_sum, 17)
            master.counts.filter(age__lt=18).delete()
            master = models.FilterSumModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_sum, 17)
            master.counts.filter(age=19)[0].delete()
            master = models.FilterSumModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_sum, 8)
            item = master.counts.filter(age=18)[0]
            item.age = 15
            item.save()
            master = models.FilterSumModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_sum, 0)
            item = master.counts.filter(age=15)[0]
            item.age = 18
            item.save()
            master = models.FilterSumModel.objects.get(pk=master.pk)
            self.assertEqual(master.active_item_sum, 8)


class CommandsTestCase(TransactionTestCase):
    @patch("select.select")
    @patch("denorm.tasks.flush_via_queue")
    def test_denorm_queue(self, flush_via_queue, select):
        "Test denorm_queue command."
        call_command("denorm_queue", run_once=True)
        select.assert_called_once()
        flush_via_queue.delay.assert_called_once()

    def test_makemigrations(self):
        "Test makemigrations command."
        call_command("makemigrations", verbosity=0)

    def test_denorm_init(self):
        "Test denorm_init command."
        call_command("denorm_init")

    def test_denorm_drop(self):
        "Test denorm_init command."
        call_command("denorm_drop")

    def test_denorm_flush(self):
        "Test denorm_init command."
        call_command("denorm_flush")

    def test_denorm_rebuild(self):
        "Test denorm_init command."
        call_command("denorm_rebuild")

    def test_denorm_sql(self):
        "Test denorm_init command."
        import sys

        sys.stdout = StringIO()

        call_command("denorm_sql")

        sys.stdout = sys.__stdout__



# Generated by Django 4.2.23 on 2025-08-18 09:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("denorm", "0014_parametrized_notify"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="dirtyinstance",
            name="processing_finished",
        ),
        migrations.RemoveField(
            model_name="dirtyinstance",
            name="processing_started",
        ),
        migrations.RemoveField(
            model_name="dirtyinstance",
            name="success",
        ),
        migrations.RemoveField(
            model_name="dirtyinstance",
            name="traceback",
        ),
    ]



from django.conf import settings

DENORM_DISABLE_AUTOTIME_DURING_FLUSH = getattr(
    settings, "DENORM_DISABLE_AUTOTIME_DURING_FLUSH", False
)

DENORM_AUTOTIME_FIELD_NAMES = getattr(settings, "DENORM_AUTOTIME_FIELD_NAMES", [])

DENORM_BATCH_SIZE = getattr(settings, "DENORM_BATCH_SIZE", 5000)



from django.core.management.base import BaseCommand

from denorm import denorms


class Command(BaseCommand):
    help = "Recalculates the value of every single denormalized model field in the whole project."

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-flush",
            action="store_true",
            help="Specify this if you alredy run denorm_queue in background",
        )
        parser.add_argument("--model-name", type=str, default=None)

    def handle(self, no_flush, model_name, *args, **kwargs):
        verbosity = int((kwargs.get("verbosity", 0)))
        denorms.rebuildall(
            verbose=verbosity > 1,
            model_name=model_name,
            flush_=not no_flush,
        )



# Generated by Django 3.2.7 on 2021-10-03 06:18

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    replaces = [
        ("denorm", "0001_initial"),
        ("denorm", "0002_dirtyinstance_func_name"),
        ("denorm", "0003_auto_20211002_1955"),
        ("denorm", "0004_alter_dirtyinstance_success"),
        ("denorm", "0005_dirtyinstance_created_on"),
        ("denorm", "0006_auto_20211003_0346"),
        ("denorm", "0007_auto_created_on_now"),
        ("denorm", "0008_alter_dirtyinstance_object_id"),
        ("denorm", "0009_alter_dirtyinstance_object_id"),
        ("denorm", "0010_alter_dirtyinstance_object_id"),
        ("denorm", "0011_alter_dirtyinstance_object_id"),
        ("denorm", "0012_alter_dirtyinstance_object_id"),
    ]

    initial = True

    dependencies = [
        ("contenttypes", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="DirtyInstance",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.TextField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
        migrations.RunSQL(
            sql="\n        -- Notify when record get inserted into 'django_denorm' table\n"
            "        CREATE OR REPLACE FUNCTION notify_django_denorm_queue()\n"
            "          RETURNS trigger AS $$\n        DECLARE\n        BEGIN\n"
            "          PERFORM pg_notify('django_denorm_process', '');\n          RETURN NEW;\n"
            "        END;\n        $$ LANGUAGE plpgsql;\n\n        "
            "CREATE TRIGGER notify_django_denorm_queue\n          AFTER INSERT ON denorm_dirtyinstance\n"
            "          FOR EACH STATEMENT\n          EXECUTE PROCEDURE notify_django_denorm_queue();\n"
            "        ",
            reverse_sql="\n            DROP TRIGGER notify_django_denorm_queue ON denorm_dirtyinstance;\n     "
            "       DROP FUNCTION notify_django_denorm_queue;\n            ",
        ),
        migrations.AddField(
            model_name="dirtyinstance",
            name="processing_finished",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="dirtyinstance",
            name="traceback",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="dirtyinstance",
            name="created_on",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AddField(
            model_name="dirtyinstance",
            name="func_name",
            field=models.TextField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.IntegerField(db_index=True, default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="dirtyinstance",
            name="processing_started",
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name="dirtyinstance",
            name="success",
            field=models.NullBooleanField(db_index=True),
        ),
        migrations.RunSQL(
            sql="ALTER TABLE denorm_dirtyinstance ALTER COLUMN created_on SET DEFAULT CURRENT_TIMESTAMP",
            reverse_sql="",
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.2.3 on 2023-07-09 20:28

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0007_alter_post_response_count"),
    ]

    operations = [
        migrations.AlterField(
            model_name="attachment",
            name="forum",
            field=models.ForeignKey(
                blank=True,
                default=None,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="test_app.forum",
            ),
        ),
    ]



import logging
import select
import sys

import psycopg2.extensions
from django.core.management.base import BaseCommand
from django.db import connection

from denorm.db import const
from denorm.tasks import flush_via_queue

logger = logging.getLogger(__name__)


class Command(BaseCommand):

    help = "Runs a process that checks for dirty fields and updates them in regular intervals."

    def add_arguments(self, parser):
        parser.add_argument(
            "--run-once",
            action="store_true",
            help="Used for testing. Causes event loop to run once. ",
        )

    def handle(self, run_once=False, **options):
        ran_once = False

        crs = (
            connection.cursor()
        )  # get the cursor and establish the connection.connection
        pg_con = connection.connection
        pg_con.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        crs.execute(f"LISTEN {const.DENORM_QUEUE_NAME}")

        logger.info(
            f"waiting for notifications on channel '{const.DENORM_QUEUE_NAME}'..."
        )
        while True:
            if ran_once and run_once:
                break
            ran_once = True

            try:
                if select.select([pg_con], [], [], None) == ([], [], []):
                    logger.warning("timeout")
                else:
                    pg_con.poll()
                    flush_via_queue.delay()

            except KeyboardInterrupt:
                sys.exit()



# Generated by Django 3.2.7 on 2021-10-02 17:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("denorm", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="dirtyinstance",
            name="func_name",
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.2.3 on 2023-07-09 20:23

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0005_skipcommentwithonly"),
    ]

    operations = [
        migrations.AlterField(
            model_name="forum",
            name="author_names",
            field=models.CharField(default="", editable=False, max_length=255),
        ),
    ]



# Generated by Django 3.2.7 on 2021-10-02 01:13

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("test_app", "0004_auto_20160306_1822"),
    ]

    operations = [
        migrations.CreateModel(
            name="SkipCommentWithOnly",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.TextField()),
                ("created_on", models.DateTimeField(auto_now_add=True)),
                ("updated_on", models.DateTimeField(auto_now=True, null=True)),
                ("post_text", models.TextField(editable=False)),
                (
                    "post",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="test_app.skippost",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import denorm.fields


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0002_auto_20151014_1049'),
    ]

    operations = [
        migrations.CreateModel(
            name='CallCounter',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', auto_created=True, primary_key=True)),
                ('called_count', models.IntegerField(editable=False)),
            ],
        ),
        migrations.CreateModel(
            name='PostExtend',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', auto_created=True, primary_key=True)),
                ('author_name', models.CharField(editable=False, max_length=255)),
                ('post', models.OneToOneField(to='test_app.Post', on_delete=models.CASCADE)),
            ],
        ),
        migrations.AlterField(
            model_name='forum',
            name='authors',
            field=models.ManyToManyField(editable=False, blank=True, to='test_app.Member'),
        ),
        migrations.AlterField(
            model_name='forum',
            name='post_count',
            field=denorm.fields.CountField('post_set', default=0),
        ),
        migrations.CreateModel(
            name='CallCounterProxy',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('test_app.callcounter',),
        ),
    ]



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_migrate


def denorm_install_triggers_after_migrate(using, **kwargs):
    from denorm import denorms

    denorms.install_triggers(using=using)


class DenormAppConfig(AppConfig):
    name = "denorm"

    def ready(self):
        if getattr(settings, "DENORM_INSTALL_TRIGGERS_AFTER_MIGRATE", True):
            post_migrate.connect(denorm_install_triggers_after_migrate, sender=self)



from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import connection, connections, models

from ..helpers import remote_field_model


class RandomBigInt(object):
    def sql(self):
        raise NotImplementedError


class TriggerNestedSelect:
    def __init__(self, table, columns, **kwargs):
        self.table = table
        self.columns = ", ".join(columns)
        self.kwargs = kwargs

    def sql(self):
        raise NotImplementedError


class TriggerAction(object):
    def __init__(self):
        pass

    def sql(self):
        pass


class TriggerActionInsert(TriggerAction):
    def __init__(self, model, columns, values):
        self.model = model
        self.columns = columns
        self.values = values

    def __repr__(self):
        return (
            f"<TriggerActionInsert object at {id(self):#02x} "
            f"INSERT INTO {self.model}({self.columns}) values({self.values})>"
        )

    def sql(self):
        raise NotImplementedError


class TriggerActionUpdate(TriggerAction):
    def __init__(self, model, columns, values, where):
        self.model = model
        self.columns = columns
        self.where = where

        self.values = []
        for value in values:
            if hasattr(value, "sql"):
                self.values.append(value.sql())
            else:
                self.values.append(value)

    def sql(self):
        raise NotImplementedError


def get_fields_with_model(model, meta):
    try:
        return [
            (f, f.model if f.model != model else None)
            for f in meta.get_fields()
            if not f.is_relation or f.one_to_one or (f.many_to_one and f.related_model)
        ]
    except AttributeError:
        return meta.get_fields_with_model()


class Trigger(object):
    def __repr__(self):
        return (
            f"{object.__repr__(self)} for model {self.subject}, {self.time} {self.event} "
            f"perform {[str(action) for action in self.actions]}"
        )

    def __init__(
        self,
        subject,
        time,
        event,
        actions,
        content_type,
        using=None,
        skip=None,
        only=None,
        func=None,
    ):
        self.subject = subject
        self.time = time
        self.event = event
        self.content_type = content_type
        self.content_type_field = None
        self.actions = []
        self.append(actions)
        self.using = using
        self.func = func

        if self.using:
            self.connection = connections[self.using]
        else:
            self.connection = connection

        from django.contrib.contenttypes.fields import GenericRelation

        fields_with_model = None

        if isinstance(subject, models.ManyToManyField):
            self.model = None
            self.db_table = subject.m2m_db_table()
            self.fields = [
                (subject.m2m_column_name(), ""),
                (subject.m2m_reverse_name(), ""),
            ]

        elif isinstance(subject, GenericRelation):
            self.model = None
            self.db_table = remote_field_model(subject)._meta.db_table
            self.fields = [
                (k.attname, k.db_type(connection=self.connection))
                for k, v in get_fields_with_model(
                    remote_field_model(subject), remote_field_model(subject)._meta
                )
                if not v
            ]
            self.content_type_field = subject.content_type_field_name + "_id"

        elif isinstance(subject, models.ForeignKey):
            self.model = subject.model
            self.db_table = self.model._meta.db_table
            skip = (skip or ()) + getattr(self.model, "denorm_always_skip", ())
            only = (only or ()) + getattr(self.model, "denorm_always_only", ())
            fields_with_model = get_fields_with_model(subject.model, self.model._meta)
            field_names_with_model = [k.attname for k, v in fields_with_model]

            if only:
                # ``only`` parametr was given and it is expected to be an iterable of field
                # names, which will be watched for updates (an opposite to ``skip`` parameter).
                # Let's see if every single field name passed actually exists:
                for field_name in only:
                    if field_name not in field_names_with_model:
                        raise FieldDoesNotExist(
                            f'Field name "{field_name}", passed as a parameter to "only" to a trigger for class or '
                            f'included in class\' "denorm_always_only" for class {self.subject}, '
                            f'does not exist. Field names available: {", ".join(field_names_with_model)}'
                        )
            else:
                # not only
                only = field_names_with_model

            self.fields = [
                (k.attname, k.db_type(connection=self.connection))
                for k, v in fields_with_model
                if not v and k.attname not in skip and k.attname in only
            ]

        elif hasattr(subject, "_meta"):
            self.model = subject
            self.db_table = self.model._meta.db_table
            # FIXME: need to check get_parent_list and add triggers to those
            # The below will only check the fields on *this* model, not parents
            skip = (skip or ()) + getattr(self.model, "denorm_always_skip", ())
            only = (only or ()) + getattr(self.model, "denorm_always_only", ())

            self.fields = []

            from django.db.models.fields.related import ForeignObjectRel

            fields_with_model = get_fields_with_model(subject, self.model._meta)
            field_names_with_model = [
                k.attname
                for k, v in fields_with_model
                if not isinstance(k, ForeignObjectRel)
            ]
            if only:
                # ``only`` parametr was given and it is expected to be an iterable of field
                # names, which will be watched for updates (an opposite to ``skip`` parameter).
                # Let's see if every single field name passed actually exists:
                for field_name in only:
                    if field_name not in field_names_with_model:
                        raise FieldDoesNotExist(
                            f'Field name "{field_name}", passed as a parameter to "only" to a trigger for class or '
                            f'included in class\' "denorm_always_only" for class {self.subject}, '
                            f'does not exist. Field names available: {", ".join(field_names_with_model)}'
                        )
            else:
                # not only
                only = field_names_with_model

            for k, v in fields_with_model:
                if isinstance(k, ForeignObjectRel):
                    pass
                else:
                    if not v and k.attname not in skip and k.attname in only:

                        self.fields.append(
                            (k.attname, k.db_type(connection=self.connection))
                        )

        else:
            raise NotImplementedError

        if skip and fields_with_model:
            field_names = [k.attname for k, v in fields_with_model]
            # Check all fields given as a parameter to skip or denorm_always_skip
            # to see if they exist:
            for field_name in skip:
                if field_name not in field_names:
                    raise FieldDoesNotExist(
                        f'Field name "{field_name}", passed as a parameter to "skip" to a trigger for class or '
                        f'included in class\' "denorm_always_skip" for class {self.subject}, '
                        f'does not exist. Field names available: {", ".join(field_names)}'
                    )

        # Check if any fields left:
        if not self.fields:
            raise ImproperlyConfigured(
                f"It looks like there are no field names defined for a trigger defined for class {self.subject}. "
                f'This situation could occur if you excluded all of the fields using "only" parameter. '
                f'Please give less broad range of fields to "only" or use "skip" instead. '
            )

    def append(self, actions):
        if not isinstance(actions, list):
            actions = [actions]

        for action in actions:
            self.actions.append(action)

    def name(self):
        return "_".join(
            ["d", self.time[:3], "row", self.event[:3], "on", self.db_table]
        )

    def sql(self):
        raise NotImplementedError


class TriggerSet(object):
    def __init__(self, using=None):
        self.using = using
        self.triggers = {}
        if self.using:
            self.connection = connections[self.using]
        else:
            self.connection = connection

    def cursor(self):
        return self.connection.cursor()

    def append(self, triggers):
        if not isinstance(triggers, list):
            triggers = [triggers]

        for trigger in triggers:
            name = trigger.name()
            if name in self.triggers:
                self.triggers[name].append(trigger.actions)
            else:
                self.triggers[name] = trigger

    def install(self):
        raise NotImplementedError

    def drop(self):
        raise NotImplementedError



import time

from django.core.management.base import BaseCommand
from tqdm import tqdm

from denorm.models import DirtyInstance
from denorm.tasks import flush_via_queue


class Command(BaseCommand):
    help = (
        "Recalculates the value of every denormalized field that was marked dirty using "
        "Celery queues."
    )

    def handle(self, **kwargs):
        # Get the total count before starting
        total_count = DirtyInstance.objects.count()

        if total_count == 0:
            self.stdout.write(self.style.SUCCESS("No dirty instances to flush."))
            return

        self.stdout.write(f"Flushing {total_count} dirty instances...")

        # Start the flush task
        result = flush_via_queue.apply_async()

        # Wait for the main task to spawn all subtasks
        time.sleep(0.5)

        # Get the group result
        group_result = result.get()

        if group_result is None:
            self.stdout.write(self.style.SUCCESS("No tasks to process."))
            return

        # Create progress bar
        with tqdm(total=total_count, desc="Flushing", unit="task") as pbar:
            while not group_result.ready():
                # Count completed tasks
                completed = group_result.completed_count()
                pbar.n = completed
                pbar.refresh()
                time.sleep(0.1)

            # Final update
            pbar.n = total_count
            pbar.refresh()

        self.stdout.write(
            self.style.SUCCESS(f"Successfully flushed {total_count} dirty instances.")
        )



from datetime import timedelta

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

DEFAULT_TIMEOUT = timedelta(minutes=5)

WEEK_AGO = timedelta(days=7)


class DirtyInstance(models.Model):
    """
    Holds a reference to a model instance that may contain inconsistent data
    that needs to be recalculated.
    DirtyInstance instances are created by the insert/update/delete triggers
    when related objects change.
    """

    class Meta:
        app_label = "denorm"

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    # null=True for object_id is intentional, it is for some weird linked foreign keys
    object_id = models.IntegerField(null=True, blank=True)

    content_object = GenericForeignKey()

    func_name = models.TextField(blank=True, null=True, db_index=True)

    created_on = models.DateTimeField(auto_now_add=True, db_index=True)

    def __str__(self):
        ret = f"DirtyInstance: {self.content_type}, {self.object_id}"
        ret += f", {self.func_name=}"
        ret += f", {self.created_on=}. "
        return ret

    def content_object_for_update(self):
        """Returns a self.content_object, only locked for update. Needs
        to run inside a transaciton."""
        klass = self.content_type.model_class()
        try:
            return klass.objects.select_for_update().get(pk=self.object_id)
        except klass.DoesNotExist:
            return

    def find_similar(self, **kwargs):
        """Find similar objects to this one. Same content_type, same object_id; func_name if this
        object has func_name, but in case of no func name -- find all objects, as no func name
        means even broader scope:"""
        return DirtyInstance.objects.filter(
            content_type=self.content_type, object_id=self.object_id, **kwargs
        )

    def delete_similar(self):
        """Remove similar DirtyInstances from db, which we haven't yet processed"""
        self.find_similar().select_for_update(skip_locked=True).delete()

    def delete_this_and_similar(self):
        self.delete_similar()
        return self.delete()



# Generated by Django 3.2.7 on 2021-10-03 03:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("denorm", "0005_dirtyinstance_created_on"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dirtyinstance",
            name="created_on",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="func_name",
            field=models.TextField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.IntegerField(db_index=True, default=0),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="processing_started",
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name="dirtyinstance",
            name="success",
            field=models.NullBooleanField(db_index=True),
        ),
    ]



# Generated by Django 3.2.7 on 2021-10-03 04:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("denorm", "0011_alter_dirtyinstance_object_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dirtyinstance",
            name="object_id",
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Django settings for test_project project.
from decimal import Decimal

import django

DEBUG = True
TEMPLATE_DEBUG = DEBUG
# TEST_RUNNER = "djangosanetesting.testrunner.DstNoseTestSuiteRunner"
if Decimal(".".join([str(i) for i in django.VERSION[:2]])) >= Decimal("1.6"):
    TEST_RUNNER = "django.test.runner.DiscoverRunner"

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# On Unix systems, a value of None will cause Django to use the same
# timezone as the operating system.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = "America/Chicago"

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = "en-us"

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale
USE_L10N = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/media/"
MEDIA_ROOT = ""

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://media.lawrence.com/media/", "http://example.com/media/"
MEDIA_URL = ""

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATIC_ROOT = ""

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = "/static/"

# URL prefix for admin static files -- CSS, JavaScript and images.
# Make sure to use a trailing slash.
# Examples: "http://foo.com/static/admin/", "/static/admin/".
ADMIN_MEDIA_PREFIX = "/static/admin/"

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    #    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = "#54vuj)xosdlk%%a$ifuj8v-^z4fvupm2vmr_!2cno&g6_a-wg"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "OPTIONS": {
            "loaders": [
                "django.template.loaders.filesystem.Loader",
                "django.template.loaders.app_directories.Loader",
            ],
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.request",
                "django.template.context_processors.static",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    }
]
MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.contrib.auth.middleware.SessionAuthenticationMiddleware",
)

ROOT_URLCONF = "test_denorm_project.urls"

TEMPLATE_DIRS = ()

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.admin",
    "django.contrib.admindocs",
    # 'django_nose',
    "test_app",
    "denorm",
)

# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "mail_admins": {"level": "ERROR", "class": "django.utils.log.AdminEmailHandler"}
    },
    "loggers": {
        "django.request": {
            "handlers": ["mail_admins"],
            "level": "ERROR",
            "propagate": True,
        },
    },
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

CELERY_ALWAYS_EAGER = True



from celery import group, shared_task
from celery_singleton import Singleton

from denorm import denorms


@shared_task(base=Singleton, ignore_result=False)
def flush_single(pk: int):
    from denorm.models import DirtyInstance

    try:
        res = DirtyInstance.objects.get(pk=pk)
    except DirtyInstance.DoesNotExist:
        return True
    denorms.flush_single(res.content_type_id, res.object_id, res.content_type)
    return True


@shared_task(base=Singleton, ignore_result=False)
def flush_via_queue():
    from denorm.models import DirtyInstance

    tasks = []
    for elem in DirtyInstance.objects.all():
        tasks.append(flush_single.s(pk=elem.pk))

    if tasks:
        job = group(tasks)
        return job.apply_async()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import denorm.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DenormModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.TextField()),
                ('ham', models.TextField(editable=False)),
                ('spam', models.TextField(editable=False)),
            ],
            options={
                'abstract': False,
                'swappable': 'DENORM_MODEL',
            },
        ),
        migrations.CreateModel(
            name='RealDenormModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.TextField()),
                ('ham', models.TextField(editable=False)),
                ('eggs', models.TextField(editable=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('cachekey', denorm.fields.CacheKeyField(default=0, editable=False)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='CachedModelA',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('cached_data', denorm.fields.CachedField(default=0, editable=False)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='CachedModelB',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('data', models.CharField(max_length=255)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FailingTriggersModelA',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('order', models.SmallIntegerField(default=0)),
                ('SomeWeirdName', models.CharField(max_length=255)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FailingTriggersModelB',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('SomeWeirdName', models.TextField(editable=False)),
                ('a', models.ForeignKey(to='test_app.FailingTriggersModelA', on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Forum',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('tags_string', models.TextField(editable=False)),
                ('title', models.CharField(max_length=255)),
                ('post_count', denorm.fields.CountField(b'post_set', default=0)),
                ('cachekey', denorm.fields.CacheKeyField(default=0, editable=False)),
                ('author_names', models.CharField(max_length=255, editable=False)),
                ('path', models.TextField(editable=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('first_name', models.CharField(max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('cachekey', denorm.fields.CacheKeyField(default=0, editable=False)),
                ('full_name', models.CharField(max_length=255, editable=False)),
                ('bookmark_titles', models.TextField(null=True, editable=False, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Post',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('tags_string', models.TextField(editable=False)),
                ('title', models.CharField(max_length=255, blank=True)),
                ('forum_title', models.CharField(max_length=255, editable=False)),
                ('author_name', models.CharField(max_length=255, editable=False)),
                ('response_count', models.PositiveIntegerField(editable=False)),
                ('author', models.ForeignKey(blank=True, to='test_app.Member', null=True, on_delete=models.CASCADE)),
                ('forum', models.ForeignKey(blank=True, to='test_app.Forum', null=True, on_delete=models.CASCADE)),
                ('response_to', models.ForeignKey(related_name='responses', blank=True, to='test_app.Post', null=True, on_delete=models.CASCADE)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='SkipCommentWithAttributeSkip',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.TextField()),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('updated_on', models.DateTimeField(auto_now=True, null=True)),
                ('post_text', models.TextField(editable=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='SkipCommentWithoutSkip',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.TextField()),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('updated_on', models.DateTimeField(auto_now=True, null=True)),
                ('post_text', models.TextField(editable=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='SkipCommentWithSkip',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.TextField()),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('updated_on', models.DateTimeField(auto_now=True, null=True)),
                ('post_text', models.TextField(editable=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='SkipPost',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.TextField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType', on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='skipcommentwithskip',
            name='post',
            field=models.ForeignKey(to='test_app.SkipPost', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='skipcommentwithoutskip',
            name='post',
            field=models.ForeignKey(to='test_app.SkipPost', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='skipcommentwithattributeskip',
            name='post',
            field=models.ForeignKey(to='test_app.SkipPost', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='member',
            name='bookmarks',
            field=models.ManyToManyField(to='test_app.Post', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='forum',
            name='authors',
            field=models.ManyToManyField(to='test_app.Member', null=True, editable=False, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='forum',
            name='parent_forum',
            field=models.ForeignKey(blank=True, to='test_app.Forum', null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='cachedmodela',
            name='b',
            field=models.ForeignKey(to='test_app.CachedModelB', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='attachment',
            name='forum',
            field=models.ForeignKey(blank=True, editable=False, to='test_app.Forum', null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='attachment',
            name='post',
            field=models.ForeignKey(blank=True, to='test_app.Post', null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import denorm.fields


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='FilterCountItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('active', models.BooleanField(default=False)),
                ('text', models.CharField(default=b'', max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='FilterCountModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('active_item_count', denorm.fields.CountField(b'items', default=0)),
            ],
        ),
        migrations.CreateModel(
            name='FilterSumItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('age', models.IntegerField(default=18)),
                ('active_item_count', models.PositiveIntegerField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='FilterSumModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('active_item_sum', denorm.fields.SumField(b'counts', default=0)),
            ],
        ),
        migrations.AddField(
            model_name='filtersumitem',
            name='parent',
            field=models.ForeignKey(related_name='counts', to='test_app.FilterSumModel', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='filtercountitem',
            name='parent',
            field=models.ForeignKey(related_name='items', to='test_app.FilterCountModel', on_delete=models.CASCADE),
        ),
    ]



from django.db import transaction
from django.db.backends.utils import truncate_name

from denorm.db import base, const


class RandomBigInt(base.RandomBigInt):
    def sql(self):
        return "(9223372036854775806::INT8 * ((RANDOM()-0.5)*2.0) )::INT8"


class TriggerNestedSelect(base.TriggerNestedSelect):
    def sql(self):
        columns = self.columns
        table = self.table
        where = ", ".join(["%s = %s" % (k, v) for k, v in self.kwargs.items()])
        return (
            "SELECT DISTINCT %(columns)s FROM %(table)s WHERE %(where)s" % locals(),
            tuple(),
        )


class TriggerActionInsert(base.TriggerActionInsert):
    def sql(self):
        table = self.model._meta.db_table
        columns = "(" + ", ".join(self.columns) + ")"
        params = []
        if isinstance(self.values, TriggerNestedSelect):
            sql, nested_params = self.values.sql()
            values = "(" + sql + ")"
            params.extend(nested_params)
        else:
            values = "VALUES (" + ", ".join(self.values) + ")"

        denorm_queue_name = const.DENORM_QUEUE_NAME
        sql = (
            "BEGIN\n"
            "    INSERT INTO %(table)s %(columns)s %(values)s;\n"
            "EXCEPTION WHEN unique_violation THEN\n"
            "    -- do nothing\n"
            "END"
        ) % locals()
        return sql, params


class TriggerActionUpdate(base.TriggerActionUpdate):
    def sql(self):
        table = self.model._meta.db_table
        params = []
        updates = ", ".join(
            ["%s = %s" % (k, v) for k, v in zip(self.columns, self.values)]
        )
        if isinstance(self.where, tuple):
            where, where_params = self.where
        else:
            where, where_params = self.where, []
        params.extend(where_params)
        return "UPDATE %(table)s SET %(updates)s WHERE %(where)s" % locals(), params


class Trigger(base.Trigger):
    def name(self):
        name = base.Trigger.name(self)
        if self.content_type_field:
            name += f"_{self.content_type}"
        if self.func:
            name += f'_{self.func.__qualname__.lower().replace(".", "_")}'
        return name

    def sql(self):
        qn = self.connection.ops.quote_name

        name = truncate_name(self.name(), self.connection.ops.max_name_length() - 5)
        params = []
        action_list = []
        actions_added = set()
        for a in self.actions:
            sql, action_params = a.sql()
            if sql:
                if not sql.endswith(";"):
                    sql += ";"
                action_params = tuple(action_params)
                if (sql, action_params) not in actions_added:
                    actions_added.add((sql, action_params))
                    action_list.extend(sql.split("\n"))
                    params.extend(action_params)
        table = self.db_table
        time = self.time.upper()
        event = self.event.upper()
        content_type = self.content_type
        ct_field = self.content_type_field

        conditions = []

        if event == "UPDATE":
            for field, native_type in self.fields:
                field = qn(field)
                if native_type is None:
                    # If Django didn't know what this field type should be
                    # then compare it as text - Fixes a problem of trying to
                    # compare PostGIS geometry fields.
                    conditions.append(
                        "(OLD.%(f)s::%(t)s IS DISTINCT FROM NEW.%(f)s::%(t)s)"
                        % {"f": field, "t": "text"}
                    )
                else:
                    conditions.append(
                        "(OLD.%(f)s IS DISTINCT FROM NEW.%(f)s)" % {"f": field}
                    )

            conditions = ["(%s)" % " OR ".join(conditions)]

        if ct_field:
            ct_field = qn(ct_field)
            if event == "UPDATE":
                conditions.append(
                    "(OLD.%(ctf)s = %(ct)s) OR (NEW.%(ctf)s = %(ct)s)"
                    % {"ctf": ct_field, "ct": content_type}
                )
            elif event == "INSERT":
                conditions.append("(NEW.%s = %s)" % (ct_field, content_type))
            elif event == "DELETE":
                conditions.append("(OLD.%s = %s)" % (ct_field, content_type))

        if conditions:
            cond = " AND ".join(conditions)
            actions = "\n            ".join(action_list)
            actions = (
                """IF %(cond)s THEN
            %(actions)s
        END IF;"""
                % locals()
            )
        else:
            actions = "\n        ".join(action_list)

        comment = ""
        spaces = "        "
        if self.func:
            comment = (
                f"-- This trigger was created because of @depend_on_related of {self.func.__qualname__}\n"
                f"{spaces}-- It happens {self.time.upper()} {self.event.upper()} on {self.db_table}\n"
                f"{spaces}-- This function was autogenerated by code found in {self.__class__}\n\n{spaces}"
            )

        sql = (
            """
CREATE OR REPLACE FUNCTION f_%(name)s()
    RETURNS TRIGGER AS $$
    BEGIN
        %(comment)s%(actions)s
        RETURN NULL;
    END;
$$ LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS %(name)s ON %(table)s;

CREATE TRIGGER %(name)s
    %(time)s %(event)s ON %(table)s
    FOR EACH ROW EXECUTE PROCEDURE f_%(name)s();
"""
            % locals()
        )
        return sql, params


class TriggerSet(base.TriggerSet):
    def drop_atomic(self):
        qn = self.connection.ops.quote_name
        cursor = self.cursor()
        cursor.execute(
            "SELECT pg_class.relname, pg_trigger.tgname FROM pg_trigger "
            "LEFT JOIN pg_class ON (pg_trigger.tgrelid = pg_class.oid) "
            "WHERE pg_trigger.tgname LIKE 'denorm_%%';"
        )
        for table_name, trigger_name in cursor.fetchall():
            cursor.execute(
                "DROP TRIGGER %s ON %s;" % (qn(trigger_name), qn(table_name))
            )

    def drop(self):
        try:
            with transaction.atomic():
                self.drop_atomic()
        except AttributeError:
            self.drop_atomic()
            transaction.commit_unless_managed(using=self.using)

    def install_atomic(self):
        cursor = self.cursor()
        cursor.execute(
            "SELECT lanname FROM pg_catalog.pg_language WHERE lanname ='plpgsql'"
        )
        if not cursor.fetchall():
            cursor.execute("CREATE LANGUAGE plpgsql")
        for name, trigger in self.triggers.items():
            sql, args = trigger.sql()
            cursor.execute(sql, args)

    def install(self):
        try:
            with transaction.atomic():
                self.install_atomic()
        except AttributeError:
            self.install_atomic()
            transaction.commit_unless_managed(using=self.using)



# -*- coding: utf-8 -*-
from django.db import models


def remote_field_model(field):
    if hasattr(field, 'remote_field') and field.remote_field:  # in Django>=1.9
        remote_field_model = field.remote_field.model
        if remote_field_model == 'self':
            return field.model
        return remote_field_model
    if hasattr(field, 'rel') and field.rel:
        return field.rel.to

def find_fks(from_model, to_model, fk_name=None):
    """
    Finds all ForeignKeys on 'from_model' pointing to 'to_model'.
    If 'fk_name' is given only ForeignKeys matching that name are returned.
    """
    # get all ForeignKeys
    fkeys = [x for x in from_model._meta.fields if isinstance(x, models.ForeignKey)]

    # filter out all FKs not pointing to 'to_model'
    fkeys = [x for x in fkeys if repr(remote_field_model(x)).lower() == repr(to_model).lower()]

    # if 'fk_name' was given, filter out all FKs not matching that name, leaving
    # only one (or none)
    if fk_name:
        fk_name = fk_name if isinstance(fk_name, str) else fk_name.attname
        fkeys = [x for x in fkeys if x.attname in (fk_name, fk_name + '_id')]

    return fkeys


def find_m2ms(from_model, to_model, m2m_name=None):
    """
    Finds all ManyToManyFields on 'from_model' pointing to 'to_model'.
    If 'm2m_name' is given only ManyToManyFields matching that name are returned.
    """
    # get all ManyToManyFields
    try:
        private_fields = from_model._meta.private_fields
    except:  # Django<2.0
        private_fields = from_model._meta.virtual_fields
    m2ms = list(from_model._meta.many_to_many) + private_fields

    # filter out all M2Ms not pointing to 'to_model'
    m2ms = [x for x in m2ms if repr(remote_field_model(x)).lower() == repr(to_model).lower()]

    # if 'm2m_name' was given, filter out all M2Ms not matching that name, leaving
    # only one (or none)
    if m2m_name:
        m2m_name = m2m_name if isinstance(m2m_name, str) else m2m_name.attname
        m2ms = [x for x in m2ms if x.attname == m2m_name]

    return m2ms



# Generated by Django 4.2.3 on 2023-07-09 20:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0009_alter_callcounter_called_count"),
    ]

    operations = [
        migrations.AlterField(
            model_name="callcounter",
            name="called_count",
            field=models.IntegerField(default=None, editable=False),
        ),
    ]



"""
This file attempts to automatically load the denorm backend for your chosen
database adaptor.

Currently only postgresql is supported. If your database is not detected then you can
specify the backend in your settings file:

DATABASES = {
    'default': {
        'DENORM_BACKEND': 'denorm.db.postgresql',
    }
}
"""

# Default mappings from common postgresql equivalents
from django.db import DEFAULT_DB_ALIAS, connections

DB_GUESS_MAPPING = {
    "postgresql_psycopg2": "postgresql",
}


def backend_for_dbname(db_name):
    return "denorm.db.%s" % DB_GUESS_MAPPING.get(db_name, db_name)


if "DENORM_BACKEND" in connections[DEFAULT_DB_ALIAS].settings_dict:
    backend = connections[DEFAULT_DB_ALIAS].settings_dict["DENORM_BACKEND"]
else:
    engine = connections[DEFAULT_DB_ALIAS].settings_dict["ENGINE"]
    backend = backend_for_dbname(engine.rsplit(".", 1)[1])



import abc
import logging
from itertools import islice

from django.apps import apps
from django.contrib import contenttypes
from django.core.exceptions import FieldDoesNotExist

try:
    from django.core.exceptions import FullResultSet
except ImportError:

    class FullResultSet(Exception):
        pass


from django.db import connection, connections, transaction
from django.db.models import ManyToManyField, sql
from django.db.models.aggregates import Sum
from django.db.models.manager import Manager
from django.db.models.query_utils import Q
from django.db.models.sql.compiler import SQLCompiler
from django.db.models.sql.datastructures import Join
from django.db.models.sql.query import JoinInfo, Query
from django.db.models.sql.where import WhereNode

from denorm.contextmanagers import suppress_autotime

logger = logging.getLogger(__name__)


def many_to_many_pre_save(sender, instance, **kwargs):
    """
    Updates denormalised many-to-many fields for the model
    """
    if instance.pk:
        # Need a primary key to do m2m stuff
        for m2m in sender._meta.local_many_to_many:
            # This gets us all m2m fields, so limit it to just those that are denormed
            if hasattr(m2m, "denorm"):
                # Does some extra jiggery-pokery for "through" m2m models.
                # May not work under lots of conditions.
                try:
                    remote = m2m.remote_field  # Django>=1.10
                except AttributeError:
                    remote = m2m.rel
                if hasattr(remote, "through_model"):
                    # Clear exisiting through records (bit heavy handed?)
                    kwargs = {m2m.related.var_name: instance}

                    # Can't use m2m_column_name in a filter
                    # kwargs = { m2m.m2m_column_name(): instance.pk, }
                    remote.through_model.objects.filter(**kwargs).delete()

                    values = m2m.denorm.func(instance)
                    for value in values:
                        kwargs.update({m2m.m2m_reverse_name(): value.pk})
                        remote.through_model.objects.create(**kwargs)

                else:
                    values = m2m.denorm.func(instance)
                    try:
                        getattr(instance, m2m.attname).set(values)
                    except AttributeError:  # Django<1.10
                        setattr(instance, m2m.attname, values)


def many_to_many_post_save(sender, instance, created, **kwargs):
    if created:

        def check_resave():
            for m2m in sender._meta.local_many_to_many:
                if hasattr(m2m, "denorm"):
                    return True
            return False

        if check_resave():
            instance.save()


def get_alldenorms():
    """
    Get all denormalizations.
    """
    alldenorms = []
    for model in apps.get_models(include_auto_created=True):
        if not model._meta.proxy:
            for field in model._meta.fields:
                if hasattr(field, "denorm"):
                    if not field.denorm.model._meta.swapped:
                        alldenorms.append(field.denorm)
    return alldenorms


class Denorm:
    def __init__(self, skip=None, only=None):
        self.func = None
        self.skip = skip
        self.only = only

    def get_quote_name(self, using):
        if using:
            cconnection = connections[using]
        else:
            cconnection = connection
        return cconnection.ops.quote_name

    def setup(self, **kwargs):
        """
        Adds 'self' to the global denorm list
        and connects all needed signals.
        """

    def update(self, instance):
        """
        Updates the denormalizations in all instances in the queryset 'qs'.
        """

        # Get attribute name (required for denormalising ForeignKeys)
        field = instance._meta.get_field(self.fieldname)
        attname = field.attname

        attr = getattr(instance, attname)

        # only write new values to the DB if they actually changed
        new_value = self.func(instance)

        if isinstance(attr, Manager):
            # for a many to many field the decorated
            # function should return a list of either model instances
            # or primary keys
            old_pks = {x.pk for x in attr.all()}
            new_pks = set()

            for x in new_value:
                # we need to compare sets of objects based on pk values,
                # as django lacks an identity map.
                if hasattr(x, "pk"):
                    new_pks.add(x.pk)
                else:
                    new_pks.add(x)

            if old_pks != new_pks:
                setattr(instance, attname, new_value)
                return {}

        elif attr != new_value:
            if hasattr(field, "related_field") and isinstance(
                new_value, field.related_field.model
            ):
                setattr(instance, attname, None)
                setattr(instance, field.name, new_value)
            else:
                setattr(instance, attname, new_value)
            return {field.name: new_value}

    def get_triggers(self, using):
        return []


class BaseCallbackDenorm(Denorm):
    """
    Handles the denormalization of one field, using a python function
    as a callback.
    """

    def setup(self, **kwargs):
        """
        Calls setup() on all DenormDependency resolvers
        """
        super().setup(**kwargs)

        for dependency in self.depend:
            dependency.setup(self.model)

    def get_triggers(self, using):
        """
        Creates a list of all triggers needed to keep track of changes
        to fields this denorm depends on.
        """
        trigger_list = list()

        # Get the triggers of all DenormDependency instances attached
        # to our callback.
        for dependency in self.depend:
            trigger_list += dependency.get_triggers(using=using)
        ret = trigger_list + super().get_triggers(using=using)
        return ret


class CallbackDenorm(BaseCallbackDenorm):
    """
    As above, but with extra triggers on self as described below
    """

    def get_triggers(self, using):
        qn = self.get_quote_name(using)

        content_type = str(
            contenttypes.models.ContentType.objects.get_for_model(self.model).pk
        )

        # Create a trigger that marks any updated or newly created
        # instance of the model containing the denormalized field
        # as dirty.
        # This is only really needed if the instance was changed without
        # using the ORM or if it was part of a bulk update.
        # In those cases the self_save_handler won't get called by the
        # pre_save signal, so we need to ensure flush() does this later.
        from .db import triggers
        from .models import DirtyInstance

        action = triggers.TriggerActionInsert(
            model=DirtyInstance,
            columns=("content_type_id", "object_id"),
            values=(
                content_type,
                "NEW.%s" % qn(self.model._meta.pk.get_attname_column()[1]),
            ),
        )
        trigger_list = [
            triggers.Trigger(
                self.model,
                "after",
                "update",
                [action],
                content_type,
                using,
                self.skip,
                self.only,
            ),
            triggers.Trigger(
                self.model,
                "after",
                "insert",
                [action],
                content_type,
                using,
                self.skip,
                self.only,
            ),
        ]

        return trigger_list + super().get_triggers(using=using)


class BaseCacheKeyDenorm(Denorm):
    def __init__(self, depend_on_related, *args, **kwargs):
        self.depend = depend_on_related
        super().__init__(*args, **kwargs)
        import random

        self.func = lambda o: random.randint(-9223372036854775808, 9223372036854775807)

    def setup(self, **kwargs):
        """
        Calls setup() on all DenormDependency resolvers
        """
        super().setup(**kwargs)

        for dependency in self.depend:
            dependency.setup(self.model)

    def get_triggers(self, using):
        """
        Creates a list of all triggers needed to keep track of changes
        to fields this denorm depends on.
        """
        trigger_list = list()

        # Get the triggers of all DenormDependency instances attached
        # to our callback.
        for dependency in self.depend:
            trigger_list += dependency.get_triggers(using=using)

        return trigger_list + super().get_triggers(using=using)


class CacheKeyDenorm(BaseCacheKeyDenorm):
    """
    As above, but with extra triggers on self as described below
    """

    def get_triggers(self, using):
        qn = self.get_quote_name(using)

        content_type = str(
            contenttypes.models.ContentType.objects.get_for_model(self.model).pk
        )

        # This is only really needed if the instance was changed without
        # using the ORM or if it was part of a bulk update.
        # In those cases the self_save_handler won't get called by the
        # pre_save signal
        from .db import triggers

        action = triggers.TriggerActionUpdate(
            model=self.model,
            columns=(self.fieldname,),
            values=(triggers.RandomBigInt(),),
            where="%s = NEW.%s"
            % ((qn(self.model._meta.pk.get_attname_column()[1]),) * 2),
        )
        trigger_list = [
            triggers.Trigger(
                self.model,
                "after",
                "update",
                [action],
                content_type,
                using,
                self.skip,
                self.only,
            ),
            triggers.Trigger(
                self.model,
                "after",
                "insert",
                [action],
                content_type,
                using,
                self.skip,
                self.only,
            ),
        ]

        return trigger_list + super().get_triggers(using=using)


class TriggerWhereNode(WhereNode):
    def sql_for_columns(self, data, qn, connection, internal_type=None):
        """
        Returns the SQL fragment used for the left-hand side of a column
        constraint (for example, the "T1.foo" portion in the clause
        "WHERE ... T1.foo = 6").
        """
        table_alias, name, db_type = data
        if table_alias:
            if table_alias in ("NEW", "OLD"):
                lhs = f"{table_alias}.{qn(name)}"
            else:
                lhs = f"{qn(table_alias)}.{qn(name)}"
        else:
            lhs = qn(name)
        try:
            response = connection.ops.field_cast_sql(db_type, internal_type) % lhs
        except TypeError:
            response = connection.ops.field_cast_sql(db_type) % lhs
        return response


class TriggerFilterQuery(sql.Query):
    def __init__(self, model, trigger_alias, where=TriggerWhereNode):
        super().__init__(model, where)
        self.trigger_alias = trigger_alias
        try:

            class JoinField:
                def get_joining_columns(self):
                    return None

            join = Join(None, None, None, None, JoinField(), False)
        except BaseException:
            join = JoinInfo(None, None, None, None, ((None, None),), False, None)
        self.alias_map = {trigger_alias: join}

    def get_initial_alias(self):
        return self.trigger_alias


class AggregateDenorm(Denorm):
    __metaclass__ = abc.ABCMeta

    def __init__(self, skip=None, only=None):
        self.manager = None
        self.skip = skip
        self.only = only

    def setup(self, sender, **kwargs):
        # as we connected to the ``class_prepared`` signal for any sender
        # and we only need to setup once, check if the sender is our model.
        if sender is self.model:
            super().setup(sender=sender, **kwargs)

        # related managers will only be available after both models are initialized
        # so check if its available already, and get our manager
        if not self.manager and hasattr(self.model, str(self.manager_name)):
            self.manager = getattr(self.model, self.manager_name)

    def get_related_where(self, fk_name, using, type):
        qn = self.get_quote_name(using)

        related_where = [
            "%s = %s.%s"
            % (qn(self.model._meta.pk.get_attname_column()[1]), type, qn(fk_name))
        ]
        related_query = Query(self.manager.related.model)
        for name, value in self.filter.items():
            related_query.add_q(Q(**{name: value}))
        for name, value in self.exclude.items():
            related_query.add_q(~Q(**{name: value}))
        related_query.add_extra(
            None,
            None,
            [
                "%s = %s.%s"
                % (
                    qn(self.model._meta.pk.get_attname_column()[1]),
                    type,
                    qn(self.manager.related.field.m2m_column_name()),
                )
            ],
            None,
            None,
            None,
        )
        related_query.add_count_column()
        related_query.clear_ordering(force_empty=True)
        related_query.default_cols = False
        related_filter_where, related_where_params = related_query.get_compiler(
            using=using
        ).as_sql()
        if related_filter_where is not None:
            related_where.append("(" + related_filter_where + ") > 0")
        return related_where, related_where_params

    def m2m_triggers(self, content_type, fk_name, related_field, using):
        """
        Returns triggers for m2m relation
        """
        from .db import triggers

        related_inc_where, _ = self.get_related_where(fk_name, using, "NEW")
        related_dec_where, related_where_params = self.get_related_where(
            fk_name, using, "OLD"
        )
        related_increment = triggers.TriggerActionUpdate(
            model=self.model,
            columns=(self.fieldname,),
            values=(self.get_related_increment_value(using),),
            where=(" AND ".join(related_inc_where), related_where_params),
        )
        related_decrement = triggers.TriggerActionUpdate(
            model=self.model,
            columns=(self.fieldname,),
            values=(self.get_related_decrement_value(using),),
            where=(" AND ".join(related_dec_where), related_where_params),
        )
        trigger_list = [
            triggers.Trigger(
                related_field,
                "after",
                "update",
                [related_increment, related_decrement],
                content_type,
                using,
                self.skip,
            ),
            triggers.Trigger(
                related_field,
                "after",
                "insert",
                [related_increment],
                content_type,
                using,
                self.skip,
            ),
            triggers.Trigger(
                related_field,
                "after",
                "delete",
                [related_decrement],
                content_type,
                using,
                self.skip,
            ),
        ]
        return trigger_list

    def get_triggers(self, using):
        from .db import triggers

        if using:
            cconnection = connections[using]
        else:
            cconnection = connection

        qn = self.get_quote_name(using)

        try:  # Django>=1.9
            related_field = self.manager.field
        except AttributeError:
            related_field = self.manager.related.field
        if isinstance(related_field, ManyToManyField):
            fk_name = related_field.m2m_reverse_name()
            inc_where = [
                "%(id)s IN (SELECT %(reverse_related)s FROM %(m2m_table)s WHERE %(related)s = NEW.%(id)s)"
                % {
                    "id": qn(self.model._meta.pk.get_attname_column()[0]),
                    "related": qn(related_field.m2m_column_name()),
                    "m2m_table": qn(related_field.m2m_db_table()),
                    "reverse_related": qn(fk_name),
                }
            ]
            dec_where = [action.replace("NEW.", "OLD.") for action in inc_where]
        else:
            pk_name = qn(self.model._meta.pk.get_attname_column()[1])
            fk_name = qn(related_field.attname)
            inc_where = [f"{pk_name} = NEW.{fk_name}"]
            dec_where = [f"{pk_name} = OLD.{fk_name}"]

        content_type = str(
            contenttypes.models.ContentType.objects.get_for_model(self.model).pk
        )

        if hasattr(self.manager, "field"):  # Django>=1.9
            related_model = self.manager.field.model
        else:  # Django>=1.8
            related_model = self.manager.related.related_model
        inc_query = TriggerFilterQuery(related_model, trigger_alias="NEW")
        inc_query.add_q(Q(**self.filter))
        inc_query.add_q(~Q(**self.exclude))
        qn = SQLCompiler(inc_query, cconnection, using)
        try:
            inc_filter_where, _ = inc_query.where.as_sql(qn, cconnection)
        except FullResultSet:
            inc_filter_where, _ = ("", "")

        dec_query = TriggerFilterQuery(related_model, trigger_alias="OLD")
        dec_query.add_q(Q(**self.filter))
        dec_query.add_q(~Q(**self.exclude))
        qn = SQLCompiler(dec_query, cconnection, using)
        try:
            dec_filter_where, where_params = dec_query.where.as_sql(qn, cconnection)
        except FullResultSet:
            dec_filter_where, where_params = ("", "")

        if inc_filter_where:
            inc_where.append(inc_filter_where)
        if dec_filter_where:
            dec_where.append(dec_filter_where)
            # create the triggers for the incremental updates
        increment = triggers.TriggerActionUpdate(
            model=self.model,
            columns=(self.fieldname,),
            values=(self.get_increment_value(using),),
            where=(" AND ".join(inc_where), where_params),
        )
        decrement = triggers.TriggerActionUpdate(
            model=self.model,
            columns=(self.fieldname,),
            values=(self.get_decrement_value(using),),
            where=(" AND ".join(dec_where), where_params),
        )

        trigger_list = [
            triggers.Trigger(
                related_model,
                "after",
                "update",
                [increment, decrement],
                content_type,
                using,
                self.skip,
            ),
            triggers.Trigger(
                related_model,
                "after",
                "insert",
                [increment],
                content_type,
                using,
                self.skip,
            ),
            triggers.Trigger(
                related_model,
                "after",
                "delete",
                [decrement],
                content_type,
                using,
                self.skip,
            ),
        ]
        if isinstance(related_field, ManyToManyField):
            trigger_list.extend(
                self.m2m_triggers(content_type, fk_name, related_field, using)
            )
        return trigger_list

    @abc.abstractmethod
    def get_increment_value(self, using):
        """
        Returns SQL for incrementing value
        """

    @abc.abstractmethod
    def get_decrement_value(self, using):
        """
        Returns SQL for decrementing value
        """


class SumDenorm(AggregateDenorm):
    """
    Handles denormalization of a sum field by doing incrementally updates.
    """

    def __init__(self, skip=None, field=None):
        super().__init__(skip)
        # in case we want to set the value without relying on the
        # correctness of the incremental updates we create a function that
        # calculates it from scratch.
        self.sum_field = field
        self.func = lambda obj: (
            getattr(obj, self.manager_name)
            .filter(**self.filter)
            .exclude(**self.exclude)
            .aggregate(Sum(self.sum_field))
            .values()[0]
            or 0
        )

    def get_increment_value(self, using):
        qn = self.get_quote_name(using)

        return f"{qn(self.fieldname)} + NEW.{qn(self.sum_field)}"

    def get_decrement_value(self, using):
        qn = self.get_quote_name(using)

        return f"{qn(self.fieldname)} - OLD.{qn(self.sum_field)}"

    def get_related_increment_value(self, using):
        qn = self.get_quote_name(using)

        related_query = Query(self.manager.related.model)
        related_query.add_extra(
            None,
            None,
            [
                "%s = %s.%s"
                % (
                    qn(self.model._meta.pk.get_attname_column()[1]),
                    "NEW",
                    qn(self.manager.related.field.m2m_column_name()),
                )
            ],
            None,
            None,
            None,
        )
        related_query.add_fields([self.fieldname])
        related_query.clear_ordering(force_empty=True)
        related_query.default_cols = False
        related_filter_where, related_where_params = related_query.get_compiler(
            using=using
        ).as_sql()
        return f"{qn(self.fieldname)} + ({related_filter_where})"

    def get_related_decrement_value(self, using):
        qn = self.get_quote_name(using)

        related_query = Query(self.manager.related.model)
        related_query.add_extra(
            None,
            None,
            [
                "%s = %s.%s"
                % (
                    qn(self.model._meta.pk.get_attname_column()[1]),
                    "OLD",
                    qn(self.manager.related.field.m2m_column_name()),
                )
            ],
            None,
            None,
            None,
        )
        related_query.add_fields([self.fieldname])
        related_query.clear_ordering(force_empty=True)
        related_query.default_cols = False
        related_filter_where, related_where_params = related_query.get_compiler(
            using=using
        ).as_sql()
        return f"{qn(self.fieldname)} - ({related_filter_where})"


class CountDenorm(AggregateDenorm):
    """
    Handles the denormalization of a count field by doing incrementally
    updates.
    """

    def __init__(self, skip=None, only=None):
        super().__init__(skip=skip, only=only)
        # in case we want to set the value without relying on the
        # correctness of the incremental updates we create a function that
        # calculates it from scratch.
        self.func = (
            lambda obj: getattr(obj, self.manager_name)
            .filter(**self.filter)
            .exclude(**self.exclude)
            .count()
        )

    def get_increment_value(self, using):
        qn = self.get_quote_name(using)

        return "%s + 1" % qn(self.fieldname)

    def get_decrement_value(self, using):
        qn = self.get_quote_name(using)

        return "%s - 1" % qn(self.fieldname)

    def get_related_increment_value(self, using):
        return self.get_increment_value(using)

    def get_related_decrement_value(self, using):
        return self.get_decrement_value(using)


def rebuild_instances_of(model, *args, **kwargs):
    # create DirtyInstance for all models

    from denorm.conf import settings

    from .models import DirtyInstance

    content_type = contenttypes.models.ContentType.objects.get_for_model(model)
    objs = (
        DirtyInstance(content_type=content_type, object_id=pk)
        for pk in model.objects.filter(*args, **kwargs).values_list("pk", flat=True)
    )

    while True:
        batch = list(islice(objs, settings.DENORM_BATCH_SIZE))
        if not batch:
            break
        DirtyInstance.objects.bulk_create(batch, settings.DENORM_BATCH_SIZE)


def rebuildall(model_name=None, field_name=None, verbose=False, flush_=True):
    """
    Updates all models containing denormalized fields.
    """

    alldenorms = get_alldenorms()
    models = {}
    for denorm in alldenorms:
        current_app_label = denorm.model._meta.app_label
        current_model_name = denorm.model._meta.model.__name__
        current_app_model = f"{current_app_label}.{current_model_name}"
        if model_name is None or model_name.lower() in (
            current_app_label.lower(),
            current_model_name.lower(),
            current_app_model.lower(),
        ):
            if field_name is None or field_name == denorm.fieldname:
                models.setdefault(denorm.model, []).append(denorm)

    i = 0
    for model, denorms in models.items():
        if verbose:
            for denorm in denorms:
                msg = (
                    "making dirty instances",
                    f"{i + 1}/{len(alldenorms)}",
                    denorm.fieldname,
                    "in",
                    denorm.model,
                )
                logger.info(msg)
                i += 1

        rebuild_instances_of(model)

    if flush_:
        flush(verbose)


def drop_triggers(using=None):
    from .db import triggers

    triggerset = triggers.TriggerSet(using=using)
    triggerset.drop()


def install_triggers(using=None):
    """
    Installs all required triggers in the database
    """
    build_triggerset(using=using).install()


def build_triggerset(using=None):
    from .db import triggers

    alldenorms = get_alldenorms()

    # Use a TriggerSet to ensure each event gets just one trigger
    triggerset = triggers.TriggerSet(using=using)
    for denorm in alldenorms:
        triggerset.append(denorm.get_triggers(using=using))
    return triggerset


INTERACTIVE = False


def flush_single(content_type_id, object_id, content_type=None):
    from denorm.conf import settings

    from .models import DirtyInstance

    disable_autotime_during_flush = settings.DENORM_DISABLE_AUTOTIME_DURING_FLUSH
    autotime_field_names = settings.DENORM_AUTOTIME_FIELD_NAMES

    if content_type is None:
        from django.contrib.contenttypes.models import ContentType

        content_type = ContentType.objects.get(pk=content_type_id)

    with transaction.atomic():
        res = DirtyInstance.objects.filter(
            content_type_id=content_type.pk, object_id=object_id
        ).select_for_update(skip_locked=True)

        if not res.exists():
            return

        klass = content_type.model_class()
        try:
            obj = klass.objects.select_for_update(of=("self",), skip_locked=True).get(
                pk=object_id
            )
        except klass.DoesNotExist:
            res.delete()
            return

        if obj is None:
            # nowait=True raises OperationalError when row is locked, skip this update
            return

        func_names = set(list(res.values_list("func_name", flat=True)))

        # At this point, all_func_names contains an iterable with all
        # func_names attributes requested re-indexing.
        kw = {}

        # If there's a None in it, we might as well save the whole object without any filtering.
        # If there is no None, we need to take all the functions that need rebuilding and update
        # but only those parameters.
        if None not in func_names:
            # If not, we want to pass a list of all updated fields to update_fields,
            # but first we need to check if they're callable and present on the object.
            update_fields = []
            for func_name in func_names:
                try:
                    obj._meta.get_field(func_name)
                    update_fields.append(func_name)
                except FieldDoesNotExist:
                    continue

            if update_fields:
                kw = dict(update_fields=update_fields)

        if disable_autotime_during_flush:
            with suppress_autotime(obj, autotime_field_names):
                obj.save(**kw)
        else:
            obj.save(**kw)

        res.delete()


def flush(run_once=False):
    """
    Updates all model instances marked as dirty by the DirtyInstance
    model.
    After this method finishes the DirtyInstance table is empty and
    all denormalized fields have consistent data.
    """

    # Loop until break.
    # We may need multiple passes, because an update on one instance
    # may cause an other instance to be marked dirty (dependency chains)

    # Get all dirty markers

    ran_once = False

    from .models import DirtyInstance

    while True:
        if run_once and ran_once:
            break

        processed = 0
        for content_type_id, object_id in (
            DirtyInstance.objects.all()
            .values_list("content_type_id", "object_id")
            .distinct()
        ):
            flush_single(content_type_id, object_id)
            processed += 1

        if not processed:
            return

        ran_once = True



# Please set this ONCE when starting your project. If updated, one needs to
# rebuild the whole django_denorm app, renaming this after the tables have been
# initialized is NOT supported...

DENORM_QUEUE_NAME = "django_denorm_process"



#!/usr/bin/python

from setuptools import setup

setup(
    name="django-denorm-iplweb",
    version="1.9.0",
    description="Denormalization magic for Django",
    long_description="django-denorm-iplweb is a Django application to provide "
    "automatic management of denormalized database fields.",
    author_email="michal.dtz@gmail.com",
    url="http://github.com/mpasternak/django-denorm-iplweb/",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
    ],
    packages=[
        "denorm",
        "denorm.conf",
        "denorm.db",
        "denorm.management",
        "denorm.management.commands",
        "denorm.migrations",
    ],
)



from django.conf import settings
from django.db import connection, models

try:
    from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey, GenericRelation

from django.contrib import contenttypes
from django.core.cache import cache

from denorm import CacheKeyField, CountField, cached, denormalized, depend_on_related
from denorm.fields import SumField

settings.DENORM_MODEL = "test_app.RealDenormModel"


class FailingTriggersModelA(models.Model):
    order = models.SmallIntegerField(default=0)  # Fails for SQLite
    SomeWeirdName = models.CharField(max_length=255)  # Fails for PostgreSQL


class FailingTriggersModelB(models.Model):
    a = models.ForeignKey(FailingTriggersModelA, on_delete=models.CASCADE)

    @denormalized(models.TextField)
    @depend_on_related(FailingTriggersModelA)
    def SomeWeirdName(self):
        return self.a.SomeWeirdName


class CachedModelA(models.Model):
    b = models.ForeignKey("CachedModelB", on_delete=models.CASCADE)

    @cached(cache)
    @depend_on_related("CachedModelB")
    def cached_data(self):
        return {
            "upper": self.b.data.upper(),
            "lower": self.b.data.lower(),
        }


class CachedModelB(models.Model):
    data = models.CharField(max_length=255)


class AbstractDenormModel(models.Model):
    # Skip feature test main model.
    text = models.TextField()

    @denormalized(models.TextField)
    def ham(self):
        return "Ham and %s" % self.text

    class Meta:
        abstract = True
        app_label = "test_app"


class DenormModel(AbstractDenormModel):
    @denormalized(models.TextField)
    def spam(self):
        return "Spam and %s" % self.text

    class Meta(AbstractDenormModel.Meta):
        swappable = "DENORM_MODEL"


class RealDenormModel(AbstractDenormModel):
    @denormalized(models.TextField)
    def eggs(self):
        return "Eggs and %s" % self.text

    class Meta(AbstractDenormModel.Meta):
        pass


class Tag(models.Model):
    name = models.CharField(max_length=255)

    content_type = models.ForeignKey(
        contenttypes.models.ContentType, on_delete=models.CASCADE
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")


class TaggedModel(models.Model):
    tags = GenericRelation(Tag)

    @denormalized(models.TextField)
    @depend_on_related(Tag)
    def tags_string(self):
        return ", ".join(sorted(t.name for t in self.tags.all()))

    class Meta:
        abstract = True


class Forum(TaggedModel):
    title = models.CharField(max_length=255)

    # Simple count() aggregate
    post_count = CountField("post_set")

    cachekey = CacheKeyField()
    cachekey.depend_on_related("Post")

    @denormalized(models.CharField, max_length=255)
    @depend_on_related("Post")
    def author_names(self):
        if not self.pk:
            return ""
        return ", ".join(m.author_name for m in self.post_set.all())

    @denormalized(models.ManyToManyField, "Member", blank=True)
    @depend_on_related("Post")
    def authors(self):
        return [m.author for m in self.post_set.all() if m.author]

    # let's say this forums supports subforums, sub-subforums and so forth
    # so we can test depend_on_related('self') (for tree structures).
    parent_forum = models.ForeignKey(
        "self", blank=True, null=True, on_delete=models.CASCADE
    )

    @denormalized(models.TextField)
    @depend_on_related("self", type="forward")
    def path(self):
        if self.parent_forum:
            return self.parent_forum.path + self.title + "/"
        else:
            return "/" + self.title + "/"


class Post(TaggedModel):
    forum = models.ForeignKey(Forum, blank=True, null=True, on_delete=models.CASCADE)
    author = models.ForeignKey(
        "Member", blank=True, null=True, on_delete=models.CASCADE
    )
    response_to = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="responses",
        on_delete=models.CASCADE,
    )
    title = models.CharField(max_length=255, blank=True)

    # Brings down the forum title
    @denormalized(models.CharField, max_length=255)
    @depend_on_related(Forum)
    def forum_title(self):
        return self.forum.title

    @denormalized(models.CharField, max_length=255)
    @depend_on_related("Member", foreign_key="author")
    def author_name(self):
        if self.author:
            return self.author.name
        else:
            return ""

    @denormalized(models.PositiveIntegerField, default=0)
    @depend_on_related("self", type="backward")
    def response_count(self):
        # Work around odd issue during testing with PostgresDB
        if not self.pk:
            return 0
        rcount = self.responses.count()
        rcount += sum(x.response_count for x in self.responses.all())
        return rcount


class PostExtend(models.Model):
    # Test also OneToOneField
    post = models.OneToOneField("Post", on_delete=models.CASCADE)

    @denormalized(models.CharField, max_length=255)
    @depend_on_related("Post")
    def author_name(self):
        return self.post.author.name


class Attachment(models.Model):
    forum_as_object = False

    post = models.ForeignKey(Post, blank=True, null=True, on_delete=models.CASCADE)

    cachekey = CacheKeyField()
    cachekey.depend_on_related("Post")

    @denormalized(
        models.ForeignKey,
        Forum,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        default=None,
    )
    @depend_on_related(Post)
    def forum(self):
        if self.post and self.post.forum:
            if self.forum_as_object:
                # if forum_as_object is set, return forum denorm as an object
                return self.post.forum
            else:
                # otherwise, return as a primary key
                return self.post.forum.pk
        return None


class Member(models.Model):
    first_name = models.CharField(max_length=255)
    name = models.CharField(max_length=255)
    bookmarks = models.ManyToManyField("Post", blank=True)

    cachekey = CacheKeyField()
    cachekey.depend_on_related("Post", foreign_key="bookmarks")

    @denormalized(models.CharField, max_length=255)
    def full_name(self):
        return f"{self.first_name} {self.name}"

    @denormalized(models.TextField, null=True, blank=True)
    @depend_on_related("Post", foreign_key="bookmarks")
    def bookmark_titles(self):
        if self.id:
            return "\n".join([p.title for p in self.bookmarks.all()])


class SkipPost(models.Model):
    # Skip feature test main model.
    text = models.TextField()


class CallCounter(models.Model):
    @denormalized(models.IntegerField)
    def called_count(self):
        if not self.called_count:
            return 1
        return self.called_count + 1


class CallCounterProxy(CallCounter):
    class Meta:
        proxy = True


class SkipComment(models.Model):
    post = models.ForeignKey(SkipPost, on_delete=models.CASCADE)
    text = models.TextField()
    created_on = models.DateTimeField(auto_now_add=True)
    updated_on = models.DateTimeField(auto_now=True, null=True, blank=True)

    class Meta:
        abstract = True


class SkipCommentWithoutSkip(SkipComment):
    # Skip feature test model without a skip parameter on an updatable field.
    # he updatable field will not be skipped.
    @denormalized(models.TextField)
    @depend_on_related(SkipPost)
    def post_text(self):
        return self.post.text


class SkipCommentWithSkip(SkipComment):
    # Skip feature test model with a skip parameter on an updatable field.
    @denormalized(models.TextField, skip=("updated_on",))
    @depend_on_related(SkipPost)
    def post_text(self):
        return self.post.text


class SkipCommentWithOnly(SkipComment):
    # Skip feature test model with a skip parameter on an updatable field.
    @denormalized(models.TextField, only=("text",))
    @depend_on_related(SkipPost)
    def post_text(self):
        return self.post.text


class SkipCommentWithAttributeSkip(SkipComment):
    @denormalized(models.TextField)
    @depend_on_related(SkipPost)
    def post_text(self):
        return self.post.text

    denorm_always_skip = ("updated_on",)


class Team(models.Model):
    @denormalized(models.TextField)
    @depend_on_related("Competitor")
    def user_string(self):
        if not self.pk:
            return ""
        return ", ".join(sorted(u.name for u in self.competitor_set.all()))


class Competitor(models.Model):
    name = models.TextField()
    team = models.ForeignKey(Team, on_delete=models.CASCADE)


if connection.vendor != "sqlite":

    class FilterSumModel(models.Model):
        # Simple count() aggregate
        active_item_sum = SumField(
            "counts", field="active_item_count", filter={"age__gte": 18}
        )

    class FilterSumItem(models.Model):
        parent = models.ForeignKey(
            FilterSumModel, related_name="counts", on_delete=models.CASCADE
        )
        age = models.IntegerField(default=18)
        active_item_count = models.PositiveIntegerField(default=False)

    class FilterCountModel(models.Model):
        # Simple count() aggregate
        active_item_count = CountField(
            "items", filter={"active__exact": True}, exclude={"text": ""}
        )

    class FilterCountItem(models.Model):
        parent = models.ForeignKey(
            FilterCountModel, related_name="items", on_delete=models.CASCADE
        )
        active = models.BooleanField(default=False)
        text = models.CharField(max_length=10, default="")



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-03-06 18:22
from __future__ import unicode_literals

import denorm.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0003_auto_20160218_1341'),
    ]

    operations = [
        migrations.CreateModel(
            name='Competitor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Team',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user_string', models.TextField(editable=False)),
            ],
        ),
        migrations.AlterField(
            model_name='filtercountitem',
            name='text',
            field=models.CharField(default='', max_length=10),
        ),
        migrations.AlterField(
            model_name='filtercountmodel',
            name='active_item_count',
            field=denorm.fields.CountField('items', default=0),
        ),
        migrations.AlterField(
            model_name='filtersummodel',
            name='active_item_sum',
            field=denorm.fields.SumField('counts', default=0),
        ),
        migrations.AddField(
            model_name='competitor',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='test_app.Team'),
        ),
    ]



# -*- coding: utf-8 -*-
import logging

import django
from django.db import DatabaseError

from denorm import flush

logger = logging.getLogger(__name__)


class DenormMiddleware(object):
    """
    Calls ``denorm.flush`` during the response stage of every request. If your data mostly or only changes during
    requests this should be a good idea. If you run into performance problems with this (because ``flush()`` takes
    to long to complete) you can try using a background process or handle flushing manually instead.

    As usual the order of middleware classes matters. It makes a lot of sense to put ``DenormMiddleware``
    after ``TransactionMiddleware`` in your ``MIDDLEWARE_CLASSES`` setting.
    """

    def process_response(self, request, response):
        try:
            flush()
        except DatabaseError as e:
            logger.error(e)
        return response


if django.VERSION >= (1, 10):
    from django.utils import deprecation

    class DenormMiddleware(
        deprecation.MiddlewareMixin,
        DenormMiddleware,
    ):
        pass



from .fields import cached, denormalized, CountField, CacheKeyField
from .denorms import flush, rebuildall
from .dependencies import depend_on_related

from django.conf import settings
if hasattr(settings, 'DENORM_FLUSH_AFTER_REQUEST'):
    import warnings
    warnings.warn("The DENORM_FLUSH_AFTER_REQUEST will be deprecated in favor of the new DenormMiddleware", DeprecationWarning)

if hasattr(settings, 'DENORM_FLUSH_AFTER_REQUEST') and settings.DENORM_FLUSH_AFTER_REQUEST:
    from django.core.signals import request_finished

    def do_flush(sender, **kwargs):
        flush()
    request_finished.connect(do_flush)

__all__ = ['cached', 'denormalized', 'depend_on_related', 'flush', 'rebuildall', 'CountField', 'CacheKeyField']



from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from denorm import denorms


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates a database to execute "
            'SQL into. Defaults to the "default" database.',
        )

    help = "Recreates all triggers needed by django-denorm."

    def handle(self, **options):
        using = options["database"]
        denorms.drop_triggers(using=using)
        denorms.install_triggers(using=using)
